"""A complete parser for the linear topic map notation, using SPARK.

$Id: ltm.py,v 1.5 2000/10/20 08:49:42 larsga Exp $
"""

# todo:
# - add TM4J support
# - ability to dump in LTM format!

# --- INITIAL IMPORT TRICKERY

import sys, string

from generic import GenericScanner, GenericParser
from token import Token

try:
    from net.ontopia.topicmaps.tmproc import tm, tmutils
except AttributeError, e:
    if sys.platform[ : 4] != "java":
        print "Couldn't import tmproc (%s)" % e
        sys.exit(1)
except ImportError, e:
    if sys.platform[ : 4] != "java":
        print "Couldn't import tmproc (%s)" % e
        sys.exit(1)

try:
    from net.ontopia.topicmaps.impl.basic import TopicMapManager, TopicName, Occurrence
    from net.ontopia.topicmaps.xml import TopicMapExporter
    from net.ontopia.xml import PrettyPrinter
    from java.lang import System
    from java.io import PrintStream, FileOutputStream
except ImportError:
    if sys.platform[:4] == "java":
        print "Couldn't import Java packages (%s)" % e
        sys.exit(1)
except NameError:
    if sys.platform[:4] == "java":
        print "Couldn't import Java packages (%s)" % e
        sys.exit(1)

# --- EXCEPTION CLASS

if sys.platform[ : 4] == "java":
    from java.lang import Exception

class LTMParseException(Exception):
    "Exception raised when LTM parsing or lexical errors occur."

    def __init__(self, msg, line):
        Exception.__init__(self, msg)
        self._line = line

    def get_line_number(self):
        return self._line


# --- LEXICAL INTERPRETATION

class LTMScanner(GenericScanner):
    
    def __init__(self):
        GenericScanner.__init__(self)
        self._lineno = 1
    
    def tokenize(self, input):
        self.rv = []
        GenericScanner.tokenize(self, input)
        return self.rv

    def tokenize_file(self, filename):
        # try to sniff character encoding first
        # this is an experimental undocumented extension
        # may be standardized in later versions
        first = open(filename).readline()
        encoding = None
        if first[0] == "@":
            spos = string.find(first, '"')
            lpos = string.rfind(first, '"')
            if spos != -1 and lpos != -1:
                encoding = first[spos + 1 : lpos]
                        
        # ok, read the thing, using the encoding information
        if not encoding:
            input = open(filename).read()
        elif sys.platform[ : 4] == "java":
            from java.io import File, FileInputStream, InputStreamReader
            from java.lang import String
            import jarray
            file = File(filename)
            reader = InputStreamReader(FileInputStream(file), encoding)
            chars = jarray.zeros(file.length(), 'c')
            reader.read(chars)
            input = str(String(chars))
        elif sys.version[ : 3] == "2.0":
            import codecs
            inf = codecs.open(filename, "r", encoding)
            input = inf.read()
            inf.close()
                              
        return self.tokenize(input)

    def error(self, s, pos):
        raise LTMParseException("Lexical error at position %d (line %d)" %
                                (pos, self._lineno), self._lineno)
    
    def t_whitespace(self, s):
        r' [\s]+ '
        self._lineno = self._lineno + string.count(s, "\n")

    def t_identifier(self, s):
        r' [A-Za-z_][-A-Za-z_0-9.:]* '
        self.rv.append(Token(type = 'identifier', attr = s, line = self._lineno))

    def t_string(self, s):
        r' "[^"]*" '
        self._lineno = self._lineno + string.count(s, "\n")
        self.rv.append(Token(type = 'string', attr = s[1 : -1], line = self._lineno))
    
    def t_comment(self, s):
        r' /\* ([^*]|\*[^/])* \*/ '    
        self._lineno = self._lineno + string.count(s, "\n")
        
    # Did this so that t_separator would have lower priority than t_comment
    # This solves the / problem. Should really upgrade SPARK and use a
    # subclass instead
        
    def t_default(self, s):
        r' [\[\]/:=;(){}@,] '
        self.rv.append(Token(type = s, line = self._lineno))
        
# --- PARSER

class LTMParser(GenericParser):

    def __init__(self, engine):
        GenericParser.__init__(self, 'ltm')
        self._engine = engine

    def error(self, token):
        raise LTMParseException("Syntax error at or near `%s' token (type '%s', line %d)" % (token, token.type, token.line),
                                token.line)

    # parsing methods
    
    def p_ltm(self, args):
        ''' ltm ::= charset decllist
            ltm ::= decllist '''
        return self._engine.get_tm()

    def p_charset(self, args):
        ' charset ::= @ string '
    
    def p_decllist(self, args):
        ''' decllist ::= topic decllist
            decllist ::= assoc decllist
            decllist ::= occur decllist
            decllist ::= assoc
            decllist ::= occur
            decllist ::= topic '''

    def p_topic(self, args):
        ''' topic ::= [ identifier opt_types opt_names opt_psi ] '''

        (id, types, names, psi) = args[1 : -1]
        topic = self._engine.lookup_topic(id)
        self._engine.add_name_and_types(topic, names, types)
        self._engine.set_identity(topic, psi)
        return topic

    def p_opt_types(self, args):
        ''' opt_types ::= : idlist
            opt_types ::= '''

        if len(args) > 0:
            return map(self._engine.lookup_topic, args[1])
        else:
            return []

    def p_opt_names(self, args):
        ''' opt_names ::= names opt_names
            opt_names ::=       '''
        
        if len(args) > 0:
            args[1].append(args[0])
            return args[1]
        else:
            return []

    def p_opt_psi(self, args):
        ''' opt_psi ::= @ string
            opt_psi ::= '''

        if len(args) > 0:
            return args[1].attr
        
    def p_names(self, args):
        ''' names ::= = scope string 
            names ::= = scope string ;
            names ::= = scope string ; string
            names ::= = scope string ;        ; string
            names ::= = scope string ; string ; string '''

        basename = args[2].attr
        dispname = None
        sortname = None
        scope    = args[1]

        if args.count(";") > 0 and args[4].type == "string":
            sortname = args[4].attr

        if args.count(";") == 2:
            dispname = args[-1].attr

        return self._engine.create_name(scope, basename, dispname, sortname)
        
    def p_idlist(self, args):
        ''' idlist ::= identifier idlist
            idlist ::= identifier '''
        
        if len(args) == 1:
            return [args[0]]
        else:
            args[1].append(args[0])
            return args[1]

    def p_scope(self, args):
        ''' scope ::=
            scope ::= / idlist '''
        
        if len(args) == 0:
            return []
        else:
            return map(self._engine.lookup_topic, args[1])

    def p_assoc(self, args):
        ''' assoc ::= identifier ( assoc-role assoc-role-list ) '''
        
        assoc_type = self._engine.lookup_topic(args[0])
        assoc = self._engine.create_assoc(assoc_type)

        role_list = args[3]
        role_list.append(args[2])

        for (topic, role_type) in role_list:
            self._engine.add_assoc_topic(assoc, topic, role_type)
        
        return assoc

    def p_assoc_role_list(self, args):
        ''' assoc-role-list ::= , assoc-role assoc-role-list
            assoc-role-list ::= '''
        
        if len(args) == 0:
            return []
        else:
            args[2].append(args[1])
            return args[2]
    
    def p_assoc_role(self, args):
        ''' assoc-role ::= topic
            assoc-role ::= topic : identifier '''
        
        topic = args[0]
        if len(args) > 1:
            role_type = self._engine.lookup_topic(args[2])
        else:
            role_type = None

        return (topic, role_type)

    def p_occur(self, args):
        ' occur ::= { identifier , identifier , string } '
        
        topic = self._engine.lookup_topic(args[1])
        occrole = self._engine.lookup_topic(args[3])
        loc = args[5].attr

        occ = self._engine.create_occurrence(occrole, loc)
        self._engine.add_occurrence(topic, occ)
        return occ

# --- TMPROC ADAPTER CLASS
    
class TmprocAdapter:

    def __init__(self):
        self._tm = tm.TopicMap()

    def get_tm(self):
        return self._tm
        
    def set_identity(self, topic, identity):
        topic.set_identity(identity)
        
    def add_name_and_types(self, topic, names, types):
        for name in names:
            topic.add_name(name)

        for type in types:
            topic.add_type(type)

    def add_occurrence(self, topic, occ):
        topic.add_occurrence(occ)
            
    def lookup_topic(self, sgmlid):
        sgmlid = sgmlid.attr
        topic = self._tm[sgmlid]
        if topic == None:
            topic = tm.Topic(sgmlid)
            self._tm.add_topic(topic)
        return topic

    def create_assoc(self, assoc_type):
        assoc = tm.Association(None, assoc_type)
        self._tm.add_association(assoc)
        return assoc

    def add_assoc_topic(self, assoc, topic, role_type):
        role = tm.AssociationRole(None, None, [topic])
        if role_type:
            role.set_type(role_type)
        assoc.add_role(role)

    def create_name(self, scopes, basename, dispname, sortname):
        if dispname:
            dispname = [tm.Name(dispname)]
        else:
            dispname = []

        if sortname:
            sortname = [tm.Name(sortname)]
        else:
            sortname = []
            
        return tm.TopicName(scopes, [tm.Name(basename)], dispname, sortname)

    def create_occurrence(self, occrole, loc):
        return tm.Occurrence(None, occrole, [], [tm.Locator(loc)])

    def post_process(self):
        # add missing assoc role types
        for assoc in self._tm.get_associations():
            for ar in assoc.get_roles():
                if ar.get_type() == None:
                    types = ar.get_locators()[0].get_types()
                    if types != []:
                        ar.set_type(types[0])
    
# --- ONTOPIA TOPIC MAP ENGINE ADAPTER
                        
class AtlasAdapter:

    def __init__(self):
        self._tmm = TopicMapManager()
        self._tm = self._tmm.getTopicMapFactory().createTopicMap()
        self._tmb = self._tmm.getTopicMapFactory().getTopicMapBuilder()

    def get_tm(self):
        return self._tm

    def set_identity(self, topic, identity):
        topic.setIdentity(identity)
    
    def add_name_and_types(self, topic, names, types):
        for name in names:
            topic.addName(name)

        for type in types:
            topic.addType(type)

    def add_occurrence(self, topic, occ):
        topic.addOccurrence(occ)
            
    def lookup_topic(self, sgmlid):
        sgmlid = sgmlid.attr
        
        topic = self._tm.getObjectById(sgmlid)
        if topic == None:
            topic = self._tmb.makeTopic(self._tm, sgmlid)

        return topic
    
    def create_assoc(self, assoc_type):
        assoc = self._tmb.makeAssociation(self._tm)
        assoc.setType(assoc_type)
        return assoc

    def add_assoc_topic(self, assoc, topic, role_type):
        r1 = self._tmb.makeAssociationRole(assoc)
        r1.setPlayer(topic)
        if role_type:
            r1.setType(role_type)
        return r1
    
    def create_name(self, scopes, basename, dispname, sortname):
        topic_name = TopicName(self._tmm)
        name = self._tmb.makeBaseName(topic_name)
        name.setValue(basename)

        if dispname != None:
            name = self._tmb.makeDisplayName(topic_name)
            name.setValue(dispname)

        if sortname != None:
            name = self._tmb.makeSortName(topic_name)
            name.setValue(sortname)

        for theme in scopes:
            topic_name.addTheme(theme)
            
        return topic_name

    def add_assoc_roles(self, assoc, art1, art2):
        roles = assoc.getRoles()
        ar1 = roles.get(0)
        ar2 = roles.get(1)
        
        ar1.setType(art1)
        ar2.setType(art2)

    def create_occurrence(self, occrole, loc):
        occur = Occurrence(self._tmm)
        occur.setType(occrole)
        
        locator = self._tmb.makeLocator()
        locator.setAddress(loc)
        occur.setLocator(locator)
        
        return occur

    def post_process(self):
        # add missing assoc role types
        iterator = self._tm.getAssociations().iterator()
        while iterator.hasNext():
            assoc = iterator.next()

            it2 = assoc.getRoles().iterator()
            while it2.hasNext():
                ar = it2.next()
                if ar.getType() != None:
                    it3 = ar.getTopic().getTypes().iterator()
                    if it3.hasNext():
                        ar.setType(it3.next())

# --- XML PrettyPrinter
# stolen from xmlarch

try:
    from xml.sax import saxlib
    class Prettifier(saxlib.HandlerBase):

        def __init__(self, writer=sys.stdout, encoding='iso-8859-1'):
            self.writer = writer
            self.elem_level=0
            self.indentor = "  "
            self.ignore_outside_docelem = 1
            self.elem_stack = []

            self.FULLINDENT = 0
            self.NOINDENT = 1

            self.CHILDREN = 0
            self.NOCHILDREN = 1

            self.xmldecl = '<?xml version="1.0" encoding="%s"?>\n' % encoding

        def indent(self, mode):
            if mode == self.FULLINDENT:
                self.writer.write("\n")
            elif mode == self.NOINDENT:
                return

            self.writer.write(self.indentor * self.elem_level)

        def set_xml_decl(self, xmldecl):
            self.xmldecl = xmldecl

        def set_writer(self, writer):
            self.writer = writer

        def processingInstruction (self, target, remainder):
            self.writer.write("<?" + target + " " + remainder + "?>\n")

        def startElement(self, name, amap):
            if not self.elem_stack:
                self.elem_stack.append(self.NOCHILDREN)
            else:
                self.elem_stack[-1] = self.CHILDREN
                self.indent(self.FULLINDENT)
                self.elem_stack.append(self.NOCHILDREN)

            self.writer.write("<" + name)

            a_names=amap.keys()
            a_names.sort()

            for a_name in a_names:
                self.writer.write(" " + a_name + "=\"")
                self.writer.write(amap[a_name])
                self.writer.write("\"")
            self.writer.write(">")

            self.elem_level = self.elem_level + 1

        def endElement(self, name):
            if self.elem_stack[-1] == self.CHILDREN:
                self.indent(self.FULLINDENT)

            self.writer.write("</" + name + ">")

            self.elem_level= self.elem_level - 1
            del self.elem_stack[-1]

        def ignorableWhitespace(self, data, start_ix, length):
            return

        def characters(self, data, start_ix, length):
            self.writer.write(string.strip(data[start_ix:start_ix+length]))

        def startDocument(self):
            self.writer.write(self.xmldecl)

        def endDocument(self):
            self.writer.write("\n")
except:
    if sys.platform[ : 4] != "java":
        raise
                    
# --- MAIN EXTERNAL API

def parse(filename, engine = None):
    engine = engine or create_engine()
    sc = LTMScanner()
    tokens = sc.tokenize_file(filename)

    p = LTMParser(engine)
    tm = p.parse(tokens)
    engine.post_process()

    return tm

def parse_string(string, engine = None):
    engine = engine or create_engine()
    sc = LTMScanner()
    tokens = sc.tokenize(string)

    p = LTMParser(engine)
    tm = p.parse(tokens)
    engine.post_process()

    return tm
    
def export(topicmap, out, encoding):
    if globals().has_key("tm") and isinstance(topicmap, tm.TopicMap):
        p = Prettifier(out, encoding)
        exp = tmutils.TMExporter().export(topicmap, p)
    else:
        exp = TopicMapExporter()
        exp.export(topicmap, PrettyPrinter(out, encoding), "attribute")    

def create_engine():
    if sys.platform[ : 4] == "java":
        return AtlasAdapter()
    else:
        return TmprocAdapter()

def convert(filename, out, encoding = 'iso-8859-1'):
    if sys.platform[ : 4] != "java" and sys.version[ : 3] == "2.0":
        import codecs
        out = codecs.lookup(encoding)[3](out)

    export(parse(filename), out, encoding)

# --- COMMAND-LINE APPLICATION

def _usage(error):
    print "\n  Ontopia LTM processor 1.0\n"
    print "Error: %s\n" % error
    print "Usage:"
    print "  ltm.py [<options>] <infile> [<outfile>] "
    print
    print "Options:"
    print "  --encoding <output encoding>     # ignored in Python 1.5.2!"
    print
    sys.exit(1)

def _cmd_line():
    try:
        import getopt
        (options, args) = getopt.getopt(sys.argv[1 : ], "", ["encoding="])

        if len(args) < 1 or len(args) > 2:
            _usage("Too many or too few arguments")
    except getopt.error, e:
        _usage(str(e))

    infile = args[0]
    if options != []:
        encoding = options[0][1]
    else:
        encoding = 'iso-8859-1'

    if sys.platform[ : 4] == "java":        
        if len(args) == 2:
            out = PrintStream(FileOutputStream(args[1]))
        else:
            out = System.out
    else:
        if len(args) == 2:
            out = open(args[1], "w")
        else:
            out = sys.stdout

    convert(infile, out, encoding)
    
if __name__ == "__main__":
    _cmd_line()
